# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from .._const import (
    DATETIME_FORMAT,
    CardType,
    CommonCardKey,
)
from ._base import AbstractCardDataFetcher
from ._common import (
    dump_organization,
    to_chart_data,
)


def ghc_starred_count_helper(ghc_client):
    return {"stars": ghc_client.starred_count}


def ghc_organizations_helper(user):
    return {"organizations": [
        dump_organization(organization) for organization in user.get_orgs()]}


def ghc_languages_helper(user):
    language_mapping = {}
    for repo in user.get_repos():
        language_mapping[repo.language] = language_mapping.get(
            repo.language, 0) + 1

    try:
        del language_mapping[None]
    except KeyError:
        pass

    return {"languages": to_chart_data(language_mapping, 5)}


class UserCardDataFetcher(AbstractCardDataFetcher):

    @property
    def type(self):
        return CardType.USER

    def fetch(self):
        self._logger.debug("fetching user data: id={}".format(self.id))

        card_data = super(UserCardDataFetcher, self).fetch()
        user = self._pygh_client.get_user(self.id)

        thread_list = [
            self._pool.apply_async(
                ghc_starred_count_helper, args=[self._ghc_client]),
            self._pool.apply_async(
                ghc_organizations_helper, args=[user]),
            self._pool.apply_async(
                ghc_languages_helper, args=[user]),
        ]

        # this will raise UnknownObjectException when failed to get data
        card_data["profile_name"] = user.name

        card_data[CommonCardKey.AVATAR_URL] = user.avatar_url
        card_data[CommonCardKey.CARD_TYPE] = CardType.USER
        card_data[CommonCardKey.CREATED_AT] = user.created_at.strftime(
            DATETIME_FORMAT)
        card_data[CommonCardKey.DESCRIPTION] = user.bio
        card_data[CommonCardKey.EMOJIS] = self._get_emoji_mapping(user.bio)
        card_data[CommonCardKey.HTML_URL] = user.html_url
        card_data[CommonCardKey.NAME] = self.id
        card_data[CommonCardKey.UPDATED_AT] = user.updated_at.strftime(
            DATETIME_FORMAT)
        card_data["blog"] = user.blog
        card_data["company"] = user.company
        card_data["email"] = user.email
        card_data["followers"] = user.followers
        card_data["following"] = user.following
        card_data["location"] = user.location
        card_data["public_gists"] = user.public_gists
        card_data["public_repos"] = user.public_repos

        for i, thread in enumerate(thread_list):
            thead_id = "thread {:d}/{:d}".format(i + 1, len(thread_list))
            self._logger.debug("wait for {}".format(thead_id))
            card_data.update(thread.get())
            self._logger.debug("complete {}".format(thead_id))

        return card_data



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

from ._organization import OrganizationCardDataFetcher
from ._repository import RepositoryCardDataFetcher
from ._user import UserCardDataFetcher



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import print_function
from __future__ import unicode_literals

import github
from github.GithubException import UnknownObjectException
import logbook
import pytest

from cli._const import CardType
from cli._detector import GithubIdDetector


def monkey_get_organization(a, b):
    raise UnknownObjectException("dummy", "dummy")


logger = logbook.Logger("test")


class Test_GithubIdDetector_constructor(object):

    @pytest.mark.parametrize(["value", "expected"], [
        [None, ValueError],
        ["", ValueError],
        ["/", ValueError],
        ["//", ValueError],
        ["a/b/c", ValueError],
    ])
    def test_exception(self, value, expected):
        with pytest.raises(expected):
            GithubIdDetector(value, logger, pygh_client=github.Github(None))


class Test_GithubIdDetector_id(object):

    @pytest.mark.parametrize(["value", "expected"], [
        ["thombashi", "thombashi"],
        ["thombashi/", "thombashi"],
        ["/thombashi", "thombashi"],
        ["/thombashi/", "thombashi"],
        ["thombashi/ghscard", "thombashi/ghscard"],
        [" thombashi / ghscard ", "thombashi/ghscard"],
    ])
    def test_normal(self, monkeypatch, value, expected):
        monkeypatch.setattr(
            github.Github, "get_organization", monkey_get_organization)

        detector = GithubIdDetector(
            value, logger, pygh_client=github.Github(None))

        assert detector.id == expected


class Test_GithubIdDetector_is_user(object):

    @pytest.mark.parametrize(["value", "expected"], [
        ["thombashi", True],
        ["thombashi/", True],
        ["/thombashi", True],
        ["/thombashi/", True],
    ])
    def test_normal(self, monkeypatch, value, expected):
        monkeypatch.setattr(
            github.Github, "get_organization", monkey_get_organization)

        detector = GithubIdDetector(
            value, logger, pygh_client=github.Github(None))

        assert detector.is_user() == expected


class Test_GithubIdDetector_is_organization(object):

    @pytest.mark.parametrize(["value", "expected"], [
        ["thombashi", True],
        ["thombashi/", True],
        ["/thombashi", True],
        ["/thombashi/", True],
    ])
    def test_normal(self, monkeypatch, value, expected):
        monkeypatch.setattr(
            github.Github, "get_organization", lambda a, b: a)

        detector = GithubIdDetector(
            value, logger, pygh_client=github.Github(None))

        assert detector.is_organization() == expected


class Test_GithubIdDetector_is_repository(object):

    @pytest.mark.parametrize(["value", "expected"], [
        ["thombashi/ghscard", True],
        [" thombashi / ghscard ", True],
    ])
    def test_normal(self, value, expected):
        detector = GithubIdDetector(
            value, logger, pygh_client=github.Github(None))

        assert detector.is_repository() == expected


class Test_GithubIdDetector_get_id_type(object):

    @pytest.mark.parametrize(["value", "expected"], [
        ["thombashi", CardType.USER],
        ["thombashi/ghscard", CardType.REPOSITORY],
    ])
    def test_normal(self, monkeypatch, value, expected):
        monkeypatch.setattr(
            github.Github, "get_organization", monkey_get_organization)

        detector = GithubIdDetector(
            value, logger, pygh_client=github.Github(None))

        assert detector.get_id_type() == expected



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from .._const import CommonCardKey


def dump_organization(organization):
    if not organization:
        return {}

    if organization.name:
        organization_name = organization.name
    else:
        organization_name = organization.html_url.split("/")[-1]

    return {
        CommonCardKey.AVATAR_URL: organization.avatar_url,
        CommonCardKey.HTML_URL: organization.html_url,
        CommonCardKey.NAME: organization_name,
        "public_repos": organization.public_repos,
    }


def to_chart_data(label_count_mapping, aggregate_threshold):
    if not label_count_mapping:
        return {
            "labels": [],
            "data": [],
        }

    label_name_list = []
    label_count_list = []
    others_count = None

    for i, kv in enumerate(sorted(
            label_count_mapping.items(), key=lambda x: x[1], reverse=True)):
        key, value = kv

        if (i + 1) > aggregate_threshold:
            if not others_count:
                others_count = value
            else:
                others_count += value

            continue

        label_name_list.append(key)
        label_count_list.append(value)

    if others_count:
        label_name_list.append("others")
        label_count_list.append(others_count)

    return {
        "labels": label_name_list,
        "data": label_count_list,
    }



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import abc
import datetime
import multiprocessing

import six

from .._const import (
    CARD_DATA_VERSION,
    DATETIME_FORMAT,
    CommonCardKey,
    Result,
)
from .._error import ApiStatusError


@six.add_metaclass(abc.ABCMeta)
class AbstractCardDataFetcher(object):

    @property
    def access_token(self):
        return self.__access_token

    @abc.abstractproperty
    def type(self):   # pragma: no cover
        pass

    @property
    def id(self):
        return self.__id

    def __init__(self, pygh_client, ghc_client, id, logger):
        self.__id = id
        self._logger = logger

        self._pygh_client = pygh_client
        self._ghc_client = ghc_client

        if self._pygh_client.get_api_status() == "major":
            raise ApiStatusError("GitHub API status is in red status")

        self._pool = multiprocessing.Pool(processes=4)

    def fetch(self):
        return {
            CommonCardKey.ID: self.id,
            CommonCardKey.FETCHD_AT: datetime.datetime.now().strftime(
                DATETIME_FORMAT),
            CommonCardKey.VERSION: CARD_DATA_VERSION,
            CommonCardKey.RESULT: Result.SUCCESS,
        }

    def _get_emoji_mapping(self, text):
        try:
            emoji_list = self._ghc_client.emoji_parser.parse(text)
        except ValueError:
            emoji_list = []

        emoji_mapping = {}
        for emoji in emoji_list:
            emoji_mapping[emoji] = self._ghc_client.emoji_parser.get_url(emoji)

        return emoji_mapping



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import errno
import io
import json
import os.path
import socket

import click
import github
from github.GithubException import (
    BadCredentialsException,
    UnknownObjectException,
)
from pathvalidate import sanitize_filename
import typepy

from ._const import (
    AppConfigKey,
    CommonCardKey,
    MAX_PER_PAGE,
)
from ._detector import GithubIdDetector
from ._github_client import GitHubClient
from ._stopwatch import stopwatch
from .fetcher import (
    RepositoryCardDataFetcher,
    UserCardDataFetcher,
    OrganizationCardDataFetcher,
)


class CardGenerator(object):

    def __init__(self, logger, app_config):
        self.__logger = logger
        self.__access_token = app_config.get(
            AppConfigKey.GITHUB_API_ACCESS_TOKEN)
        self.__output_dir = app_config.get(AppConfigKey.OUTPUT_DIR)

        if typepy.is_not_null_string(self.__access_token):
            logger.debug("access token found in the configuration file")

        self.__pygh_client = github.Github(
            self.__access_token, per_page=MAX_PER_PAGE)

    def generate_card(self, github_id):
        self.__set_github_id(github_id)

        try:
            with stopwatch(self.__logger, "fetch {} {}".format(
                    github_id, self.__data_fetcher.type)):
                card_data = self.__data_fetcher.fetch()
        except socket.error as e:
            self.__logger.error("{:s}: {}".format(e.__class__.__name__, e))
            return errno.ECONNRESET
        except BadCredentialsException as e:
            self.__logger.error("invalid GitHub API public access token")
            return errno.EBADRQC
        except KeyboardInterrupt:
            self._pool.terminate()
            raise
        except UnknownObjectException as e:
            if e.status == 404:
                message = "'{}' {}".format(
                    self.__data_fetcher.id, e.data.get("message"))
            else:
                message = e.data.message
            self.__logger.error(
                "{:s} failed to get GitHub data: type={}, id={}, status={}, "
                "message={}".format(
                    e.__class__.__name__, self.__data_fetcher.type,
                    self.__data_fetcher.id, e.status, message))
            return errno.ENODATA

        card_data_text = json.dumps(card_data)

        self.__logger.debug("fetched card data: {}".format(card_data_text))

        try:
            self.__make_output_dir()
        except TypeError:
            click.echo(card_data_text)
            return 0
        except OSError as e:
            self.__logger.error("{:s}: {}".format(e.__class__.__name__, e))

            return e.args[0]

        output_path = "{:s}.json".format(os.path.join(
            self.__output_dir,
            sanitize_filename(card_data[CommonCardKey.ID], "_")))

        try:
            with io.open(output_path, "w", encoding="utf-8") as f:
                f.write(card_data_text + "\n")
        except IOError as e:
            self.__logger.error("{:s}: {}".format(e.__class__.__name__, e))
            return e.args[0]

        self.__logger.info("written {:s} data to '{:s}'".format(
            self.__detector.get_id_type().lower(), output_path))

        return 0

    def __get_data_fetcher_class(self):
        if self.__detector.is_repository():
            return RepositoryCardDataFetcher

        if self.__detector.is_user():
            return UserCardDataFetcher

        if self.__detector.is_organization():
            return OrganizationCardDataFetcher

        raise ValueError("unknown id type: {}".format(self.__detector.id))

    def __set_github_id(self, github_id):
        self.__github_id = github_id
        self.__detector = GithubIdDetector(
            self.__github_id, self.__logger, pygh_client=self.__pygh_client)
        self.__data_fetcher = self.__create_data_fetcher()

    def __create_data_fetcher(self):
        return self.__get_data_fetcher_class()(
            pygh_client=self.__pygh_client,
            ghc_client=GitHubClient(
                logger=self.__logger, github_id=self.__detector.id,
                access_token=self.__access_token),
            id=self.__detector.id,
            logger=self.__logger)

    def __make_output_dir(self):
        if os.path.isdir(self.__output_dir):
            return

        self.__logger.debug(
            "creating directory: {}".format(self.__output_dir))

        os.makedirs(self.__output_dir)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import logbook
import typepy


def get_logger(log_level, extra_name):
    logger_name_list = []
    if typepy.is_not_null_string(extra_name):
        logger_name_list.append(extra_name)

    logger = logbook.Logger(" ".join(logger_name_list))
    logger.level = log_level
    if log_level == logbook.NOTSET:
        logger.disable()

    return logger



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals


class ApiStatusError(Exception):
    """
    Raised when GitHub API is in red status.
    """



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>

winrandom
winrandom-ctypes
"""

from __future__ import unicode_literals

import io
import os.path
import sys

import setuptools


REQUIREMENT_DIR = "requirements"
ENCODING = "utf8"

with io.open("README.rst", encoding=ENCODING) as f:
    long_description = f.read()

with io.open(
        os.path.join("docs", "pages", "introduction", "summary.txt"),
        encoding=ENCODING) as f:
    summary = f.read()

with open(os.path.join(REQUIREMENT_DIR, "requirements.txt")) as f:
    install_requires = [line.strip() for line in f if line.strip()]

with open(os.path.join(REQUIREMENT_DIR, "test_requirements.txt")) as f:
    tests_requires = [line.strip() for line in f if line.strip()]

with open(os.path.join(REQUIREMENT_DIR, "docs_requirements.txt")) as f:
    docs_requires = [line.strip() for line in f if line.strip()]

needs_pytest = set(["pytest", "test", "ptr"]).intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []

PROJECT_NAME = "ghscard"

setuptools.setup(
    name=PROJECT_NAME,
    version="0.0.5",
    url="https://github.com/thombashi/{:s}".format(PROJECT_NAME),

    author="Tsuyoshi Hombashi",
    author_email="tsuyoshi.hombashi@gmail.com",
    description=summary,
    include_package_data=True,
    install_requires=install_requires,
    keywords=["GitHub", "JavaScript", "Widget"],
    license="MIT License",
    long_description=long_description,
    packages=setuptools.find_packages(exclude=["test*"]),

    setup_requires=pytest_runner,
    tests_require=tests_requires,
    extras_require={
        "test": tests_requires,
        "docs": docs_requires,
    },
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*',

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development",
        "Topic :: Software Development :: Widget Sets",
    ],
    entry_points={
        "console_scripts": [
            "ghscard=cli.ghscard:cmd",
        ],
    })



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import errno
import sys

from appconfigpy import (
    DefaultDisplayStyle,
    ConfigItem,
    ConfigManager,
)
import appconfigpy
import click
from github.GithubException import RateLimitExceededException
import logbook
import typepy

from ._const import (
    AppConfigKey,
    PROGRAM_NAME,
    PROGRAM_VERSION,
)
from ._generator import CardGenerator
from ._logger import get_logger


QUIET_LOG_LEVEL = logbook.NOTSET
CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
    obj={},
)

logbook.StderrHandler(
    level=logbook.DEBUG,
    format_string="[{record.level_name}] {record.channel}: {record.message}"
).push_application()

CONFIG_ITEM_LIST = [
    ConfigItem(
        name=AppConfigKey.GITHUB_API_ACCESS_TOKEN,
        initial_value=None,
        prompt_text="GitHub API Personal Access Token",
        default_display_style=DefaultDisplayStyle.PART_VISIBLE),
    ConfigItem(
        name=AppConfigKey.OUTPUT_DIR,
        prompt_text="Output Directory Path",
        initial_value="."),
]


class Context(object):
    LOG_LEVEL = "LOG_LEVEL"


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=PROGRAM_VERSION)
@click.option(
    "--debug", "log_level", flag_value=logbook.DEBUG,
    help=u"for debug print.")
@click.option(
    "--quiet", "log_level", flag_value=QUIET_LOG_LEVEL,
    help=u"suppress execution log messages.")
@click.pass_context
def cmd(ctx, log_level):
    ctx.obj[Context.LOG_LEVEL] = (
        logbook.INFO if log_level is None else log_level)


@cmd.command()
@click.pass_context
def configure(ctx):
    """
    Create a configuration file which includes GitHub API public access token.

    The value you provide for the GitHub API public access token written to
    the configuration file (~/.ghscard).

    Example:

        To create a new configuration:

            $ ghscard configure
            GitHub API Personal Access Token: <input access token>
    """

    appconfigpy.set_log_level(ctx.obj[Context.LOG_LEVEL])

    app_config_manager = ConfigManager(
        config_name=PROGRAM_NAME, config_item_list=CONFIG_ITEM_LIST)

    sys.exit(app_config_manager.configure())


@cmd.command()
@click.argument("id", type=str, nargs=-1)
@click.option(
    "--api-token", default=None,
    help=u"GitHub API access token.")
@click.option(
    "-o", "--output-dir", metavar="PATH", default=None,
    help=u"Output path of the SQLite database file.")
@click.pass_context
def gen(ctx, id, api_token, output_dir):
    """
    Generate a GitHub user/repository card data file.
    ID need to either <user name> or <user name>/<repository name>.

    Example:

        $ ghscard gen thombashi/ghscard
    """

    log_level = ctx.obj[Context.LOG_LEVEL]
    logger = get_logger(log_level, u"{:s} gen".format(PROGRAM_NAME))
    appconfigpy.set_log_level(log_level)

    app_config = ConfigManager(
        config_name=PROGRAM_NAME, config_item_list=CONFIG_ITEM_LIST).load()

    if typepy.is_not_null_string(output_dir):
        app_config[AppConfigKey.OUTPUT_DIR] = output_dir

    if typepy.is_not_null_string(api_token):
        app_config[AppConfigKey.GITHUB_API_ACCESS_TOKEN] = api_token

    if not id:
        logger.error(
            u"command requires at least one argument: "
            u"<user-name> or <user-name>/<repository-name>")
        sys.exit(errno.EINVAL)

    return_code_list = []
    generator = CardGenerator(logger, app_config)
    for gh_id in id:
        try:
            return_code_list.append(generator.generate_card(gh_id))
        except KeyboardInterrupt:
            sys.exit(errno.EINTR)
        except RateLimitExceededException as e:
            logger.error(e.data.get("message"))
            sys.exit(errno.ENOSR)

    if any(return_code_list):
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    cmd()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import re

from github.GithubException import RateLimitExceededException
import typepy

from ._const import MAX_PER_PAGE
from ._emoji import EmojiParser
from ._stopwatch import stopwatch


class GitHubClient(object):

    @property
    def emojis(self):
        if self.__emojis:
            return self.__emojis

        self.__emojis = self.get("/emojis")

        return self.__emojis

    @property
    def emoji_parser(self):
        if self.__emoji_parser:
            return self.__emoji_parser

        self.__emoji_parser = EmojiParser(self.emojis)

        return self.__emoji_parser

    @property
    def repo(self):
        if self.__repos:
            return self.__repos

        self.__repos = self.get(
            "/repos/{:s}".format(self.__github_id),
            headers={"accept": "application/vnd.github.drax-preview+json"})
        # get license: https://developer.github.com/v3/licenses/

        return self.__repos

    @property
    def branches_count(self):
        return self.__get_count("branches")

    @property
    def contributors_count(self):
        return self.__get_count("contributors")

    @property
    def pulls_count(self):
        return self.__get_count("pulls")

    @property
    def tags_count(self):
        return self.__get_count("tags")

    @property
    def starred_count(self):
        return self.__get_count("starred")

    def __init__(self, logger, github_id, access_token=None):
        self._logger = logger
        self.__github_id = github_id
        self.__access_token = access_token

        self.__emojis = None
        self.__emoji_parser = None
        self.__repos = None

    def get(self, operation, headers=None, params=None):
        import requests

        if not headers:
            headers = {}

        if not params:
            params = {}

        if typepy.is_not_null_string(self.__access_token):
            headers["authorization"] = "token {:s}".format(self.__access_token)

        api_url = "https://api.github.com{:s}".format(operation)
        response = requests.get(api_url, headers=headers, params=params)
        self._logger.debug("API called: {}".format(response.url))

        try:
            response_json = response.json()
        except ValueError:
            return {}

        try:
            message = response_json.get("message")
        except AttributeError:
            return response_json

        if message:
            if re.search(
                    ".* list is too large to list .* via the API", message):
                raise IOError(message)

            if all([
                    response.status_code == 403,
                    re.search("^API rate limit exceeded for ", message),
            ]):
                raise RateLimitExceededException(message)

        return response_json

    def get_page(self, operation, page):
        return self.get(
            operation,
            params={
                "per_page": str(MAX_PER_PAGE),
                "page": page,
            }
        )

    def _get_branches(self, page):
        # https://developer.github.com/v3/repos/branches/
        return self.get_page(
            "/repos/{:s}/branches".format(self.__github_id), page=page)

    def _get_contributors(self, page):
        return self.get_page(
            "/repos/{:s}/contributors".format(self.__github_id), page=page)

    def _get_pulls(self, page):
        # https://developer.github.com/v3/pulls/
        return self.get_page(
            "/repos/{:s}/pulls".format(self.__github_id), page=page)

    def _get_tags(self, page):
        # https://developer.github.com/v3/git/tags/
        return self.get_page(
            "/repos/{:s}/tags".format(self.__github_id), page=page)

    def _get_releases(self, page):
        # https://developer.github.com/v3/repos/releases/
        return self.get_page(
            "/repos/{:s}/releases".format(self.__github_id), page=page)

    def _get_starred(self, page):
        return self.get_page(
            "/users/{:s}/starred".format(self.__github_id), page=page)

    def __get_count(self, param_name):
        attr_template = "__{:s}"
        method_template = "_get_{:s}"

        try:
            count = getattr(self, attr_template.format(param_name))
            if count:
                return count
        except AttributeError:
            pass

        total_count = 0
        page = 1

        with stopwatch(self._logger, "get {:s}".format(param_name)):
            while True:
                method_name = method_template.format(param_name)
                with stopwatch(self._logger, "{:s} page {:d}".format(method_name, page)):
                    try:
                        subtotal_count = len(
                            getattr(self, method_name)(page))
                    except IOError as e:
                        self._logger.debug(
                            "{:s}: {}".format(e.__class__.__name__, e))
                        total_count = None
                        break

                if not subtotal_count:
                    break

                total_count += subtotal_count
                page += 1

            setattr(self, attr_template.format(param_name), total_count)

        return total_count
